
# TO DO: add the new print statements from explore_categorical
def explore_numeric(df, x, figsize=(6,5) ):
  """Creates a seaborn histplot and boxplot with a share x-axis,
  Prints statements about null values, cardinality, and checks for
  constant/quasi-constant features.
  Source:{PASTE IN FINAL LESSON LINK}
  """

  ## Save null value counts and percent for printing
  null_count = df[x].isna().sum()
  null_perc = null_count/len(df)* 100


  ## Making our figure with gridspec for subplots
  gridspec = {'height_ratios':[0.7,0.3]}
  fig, axes = plt.subplots(nrows=2, figsize=figsize,
                           sharex=True, gridspec_kw=gridspec)
  # Histogram on Top
  sns.histplot(data=df, x=x, ax=axes[0])

  # Boxplot on Bottom
  sns.boxplot(data=df, x=x, ax=axes[1])

  ## Adding a title
  axes[0].set_title(f"Column: {x}", fontweight='bold')

  ## Adjusting subplots to best fill Figure
  fig.tight_layout()

  # Ensure plot is shown before message
  plt.show()


  # Print null value info
  print(f"- NaN's Found: {null_count} ({round(null_perc,2)}%)")
  # Print cardinality info
  nunique = df[x].nunique()
  print(f"- Unique Values: {nunique}")


  # Get the most most common value, its count as # and as %
  most_common_val_count = df[x].value_counts(dropna=False).head(1)
  most_common_val = most_common_val_count.index[0]
  freq = most_common_val_count.values[0]
  perc_most_common = freq / len(df) * 100

  print(f"- Most common value: '{most_common_val}' occurs {freq} times ({round(perc_most_common,2)}%)")

  # print message if quasi-constant or constant (most common val more than 98% of data)
  if perc_most_common > 98:
    print(f"\n- [!] Warning: '{x}' is a constant or quasi-constant feature and should be dropped.")
  else:
    print("- Not constant or quasi-constant.")
  return fig, axes


from sklearn.model_selection import train_test_split
from sklearn.impute import SimpleImputer
from sklearn.pipeline import make_pipeline
from sklearn.preprocessing import StandardScaler, OneHotEncoder, OrdinalEncoder
from sklearn.compose import ColumnTransformer
from sklearn.compose import make_column_selector
# set the default output to pandas
from sklearn import set_config
set_config(transform_output='pandas')



def preprocessor_maker(df, target, num_imputer='mean', fill_num = 0, ord_features=[], ord_orders=[]
            , ord_imputer='constant', fill_ord='NA', cat_imputer='constant'
            ,fill_cat='NA',  random_state=42, num_scale = True, ord_scale = True, cat_scale = True , test_size=None):


    y = df[target]
    X = df.drop([target] , axis=1)
    X_train, X_test, y_train, y_test = train_test_split(X, y, random_state=random_state, test_size=test_size)

    indicator_arr = [0,0,0]

    if (len(X_train.select_dtypes('number').columns)):
        num_selector = make_column_selector(dtype_include='number')

        if num_imputer == 'constant':
            num_imputer1 = SimpleImputer(strategy='constant', fill_value = fill_num)
        else:
            num_imputer1 = SimpleImputer(strategy=num_imputer)

        if num_scale == True:
            scaler = StandardScaler()
            num_pipe = make_pipeline(num_imputer1, scaler)
        else:
            num_pipe = make_pipeline(num_imputer1)


        num_tuple = ('numeric',num_pipe, num_selector)

        indicator_arr[0] = 1



  # ================================================
    if len(ord_features):

        ord_cols = ord_features


        if ord_imputer == 'constant':
            impute_na_ord = SimpleImputer(strategy='constant', fill_value=fill_ord)
        else:
            impute_na_ord = SimpleImputer(strategy=ord_imputer)


        ord_encoder = OrdinalEncoder(categories=ord_orders)

        # # Making a final scaler to scale category #'s
        # scaler_ord = StandardScaler()

        if ord_scale == True:
            scaler_ord = StandardScaler()
            ord_pipe = make_pipeline(impute_na_ord, ord_encoder, scaler_ord)
        else:
            ord_pipe = make_pipeline(impute_na_ord, ord_encoder)

        # ord_pipe = make_pipeline(impute_na_ord, ord_encoder, scaler_ord)



        ord_tuple = ('ordinal', ord_pipe, ord_cols)
        # indicator_arr.append(ord_tuple)
        indicator_arr[1] = 1
  # ==================================================================

    # (New) Select columns with make_column_selector----------------------------------------------
    if (len(X_train.select_dtypes('object').columns)):
        # cat_selector = make_column_selector(dtype_include='object')

        if len(ord_features):
            ohe_cols = X_train.select_dtypes('object').drop(columns=ord_cols).columns


        else:
            ohe_cols = X_train.select_dtypes('object').columns


        if cat_imputer == 'constant':
            freq_imputer = SimpleImputer(strategy='constant', fill_value=fill_cat)
        else:
            freq_imputer = SimpleImputer(strategy=cat_imputer)
            ohe = OneHotEncoder(sparse_output=False, handle_unknown='ignore')


        if cat_scale == True:
            scaler_cat = StandardScaler()
            cat_pipe = make_pipeline(freq_imputer, ohe, scaler_cat)
        else:
            cat_pipe = make_pipeline(freq_imputer, ohe)


        cat_tuple = ('categorical',cat_pipe, ohe_cols)
        indicator_arr[2] = 1


  # =======================================================================


    if indicator_arr[0] == 1 and indicator_arr[1] == 1 and indicator_arr[2] == 1:
        all_tuple = [num_tuple, ord_tuple, cat_tuple]

    elif indicator_arr[0] == 1 and indicator_arr[1] == 1 and indicator_arr[2] == 0:
        all_tuple = [num_tuple, ord_tuple]

    elif indicator_arr[0] == 1 and indicator_arr[1] == 0 and indicator_arr[2] == 1:
        all_tuple = [num_tuple, cat_tuple]

    elif indicator_arr[0] == 0 and indicator_arr[1] == 1 and indicator_arr[2] == 1:
        all_tuple = [ord_tuple, cat_tuple]

    elif indicator_arr[0] == 1 and indicator_arr[1] == 0 and indicator_arr[2] == 0:
        all_tuple = [num_tuple]

    elif indicator_arr[0] == 0 and indicator_arr[1] == 0 and indicator_arr[2] == 1:
        all_tuple = [cat_tuple]

    elif indicator_arr[0] == 0 and indicator_arr[1] == 1 and indicator_arr[2] == 0:
        all_tuple = [ord_tuple]

    else:
        print('Error')



    preprocessor = ColumnTransformer(all_tuple, verbose_feature_names_out=False)

    return preprocessor, X_train, y_train, X_test, y_test







from .main import preprocessor_maker
# from .neural_networks_builder_file import neural_networks_builder
# from .KMeansCoff_file import KMeansCoff
# from .evaluate_classification_file import evaluate_classification
# from .evaluate_regression_file import evaluate_regression
# from .explore_categorical_file import explore_categorical
# from .plot_categorical_vs_target_file import plot_categorical_vs_target
# from .explore_numeric_file import explore_numeric
# from .modelSearch_file import modelSearch
# from .plot_numeric_vs_target_file import plot_numeric_vs_target
# from .plot_PCA_file import plot_PCA
# from .all_imports_file import all_imports




"""UNIVARIATE PLOTTING FUNCTIONS FOR EDA"""
# Add the print statements to the function
def explore_categorical(df, x, fillna = True, placeholder = 'MISSING',
                        figsize = (6,4), order = None):
  """Creates a seaborn countplot with the option to temporarily fill missing values
  Prints statements about null values, cardinality, and checks for
  constant/quasi-constant features.
  Source:{PASTE IN FINAL LESSON LINK}
  """
  # Make a copy of the dataframe and fillna
  temp_df = df.copy()
  # Before filling nulls, save null value counts and percent for printing
  null_count = temp_df[x].isna().sum()
  null_perc = null_count/len(temp_df)* 100
  # fillna with placeholder
  if fillna == True:
    temp_df[x] = temp_df[x].fillna(placeholder)
  # Create figure with desired figsize
  fig, ax = plt.subplots(figsize=figsize)
  # Plotting a count plot
  sns.countplot(data=temp_df, x=x, ax=ax, order=order)
  # Rotate Tick Labels for long names
  ax.set_xticklabels(ax.get_xticklabels(), rotation=45, ha='right')
  # Add a title with the feature name included
  ax.set_title(f"Column: {x}", fontweight='bold')

  # Fix layout and show plot (before print statements)
  fig.tight_layout()
  plt.show()

  # Print null value info
  print(f"- NaN's Found: {null_count} ({round(null_perc,2)}%)")
  # Print cardinality info
  nunique = temp_df[x].nunique()
  print(f"- Unique Values: {nunique}")
  # First find value counts of feature
  val_counts = temp_df[x].value_counts(dropna=False)
  # Define the most common value
  most_common_val = val_counts.index[0]
  # Define the frequency of the most common value
  freq = val_counts.values[0]
  # Calculate the percentage of the most common value
  perc_most_common = freq / len(temp_df) * 100
  # Print the results
  print(f"- Most common value: '{most_common_val}' occurs {freq} times ({round(perc_most_common,2)}%)")
  # print message if quasi-constant or constant (most common val more than 98% of data)
  if perc_most_common > 98:
    print(f"\n- [!] Warning: '{x}' is a constant or quasi-constant feature and should be dropped.")
  else:
    print("- Not constant or quasi-constant.")
  return fig, ax



import tensorflow as tf
from tensorflow.keras.models import Sequential
from tensorflow.keras.layers import Dense
from sklearn.preprocessing import LabelEncoder
from sklearn.metrics import classification_report,ConfusionMatrixDisplay

import numpy as np
import matplotlib.pyplot as plt
import pandas as pd

# Custom function for plotting each metric
def plot_history(history, figsize=(6,12), marker='o'):

    # Get list of metrics from history
    metrics = [c for c in history.history if not c.startswith('val_')]

    ## Separate row for each metric
    fig, axes = plt.subplots(nrows=len(metrics),figsize=figsize)

    # For each metric
    for i, metric_name in enumerate(metrics):

        # Get the axis for the current metric
        ax = axes[i]

        # Get metric from history.history
        metric_values = history.history[metric_name]
        # Get epochs from history
        epochs = history.epoch

        # Plot the training metric
        ax.plot(epochs, metric_values, label=metric_name, marker=marker)

        ## Check if val_{metric} exists. if so, plot:
        val_metric_name = f"val_{metric_name}"
        if val_metric_name in history.history:
            # Get validation values and plot
            metric_values = history.history[val_metric_name]
            ax.plot(epochs,metric_values,label=val_metric_name, marker=marker)

        # Final subplot adjustments
        ax.legend()
        ax.set_title(metric_name)
    fig.tight_layout()

    return fig, axes


def neural_networks_builder(preprocessor, X_train, y_train, X_test, y_test,epochs=10,
                            validation_split=.2, layers=[{'n': 10, 'activation': 'relu'}, {'n': 3, 'activation': 'sigmoid'}]):

  le = LabelEncoder()
  y_train_enc = le.fit_transform(y_train)
  y_test_enc = le.transform(y_test)

  X_train_tf = preprocessor.fit_transform(X_train)
  X_test_tf = preprocessor.transform(X_test)

  input_shape = X_train_tf.shape[1]


  # Sequential model
  model = Sequential()

  # First hidden layer
  model.add(Dense(layers[0]['n'], # How many neurons you have in your first hidden layer
                  input_dim = input_shape, # What is the shape of your input features (number of columns)
                  activation = layers[0]['activation'])) # What activation function are you using?

  for layer in layers[1:]:
    # Second hidden layer
    model.add(Dense(layer['n'], # How many neurons you have in your second hidden layer
                    activation =layer['activation'])) # What activation function are you using?


  ### Metrics are specified during the.compile step
  # Step 2: Compile
  model.compile(loss = 'bce', optimizer = 'adam'
                , metrics=['accuracy',
                          tf.keras.metrics.Recall(name='recall'),
                          tf.keras.metrics.Precision(name='precision'),
                          ])
  model.summary()


  # =======================================================plot_history start

  # =======================================================plot_history end


  # Step 3: Fit our model
  history = model.fit(X_train_tf, y_train_enc,
                      validation_split=validation_split,
                      epochs=epochs)

  plot_history(history);

  # Evaluate neural network with builtin evaluation
  result = model.evaluate(X_test_tf, y_test_enc,return_dict=True)

  # make predictions
  y_pred_test = model.predict(X_test_tf)

  # round the predictions
  y_pred_test = np.round(y_pred_test)


  print(classification_report(y_test_enc, y_pred_test))

  ConfusionMatrixDisplay.from_predictions(y_test_enc, y_pred_test, cmap='Blues',
                                        normalize='true');


  return model



# Import standard packages
import pandas as pd
import numpy as np
# setting random state for reproducibility
SEED = 42
np.random.seed(SEED)

import matplotlib.pyplot as plt
import seaborn as sns
import missingno as msno
pd.set_option('display.max_columns',100)

# set the default output to pandas
from sklearn import set_config
set_config(transform_output='pandas')

# Import modeling tools
from sklearn.model_selection import train_test_split
from sklearn.model_selection import GridSearchCV
from sklearn.model_selection import RandomizedSearchCV


from sklearn.preprocessing import StandardScaler, OneHotEncoder, OrdinalEncoder

from sklearn.pipeline import Pipeline, make_pipeline

from sklearn.impute import SimpleImputer


from sklearn.metrics import mean_absolute_error, mean_squared_error, r2_score
from sklearn.metrics import f1_score, recall_score, accuracy_score, precision_score
from sklearn.metrics import confusion_matrix, ConfusionMatrixDisplay ,classification_report

from sklearn.linear_model import LinearRegression
from sklearn.linear_model import LogisticRegression # for classModel


from sklearn.tree import DecisionTreeRegressor
from sklearn.tree import DecisionTreeClassifier

from sklearn.neighbors import KNeighborsClassifier

from sklearn.ensemble import BaggingRegressor
from sklearn.ensemble import RandomForestRegressor
from sklearn.ensemble import BaggingClassifier
from sklearn.ensemble import RandomForestClassifier

from sklearn.compose import ColumnTransformer
from sklearn.compose import make_column_selector


import os
import joblib


from imblearn.over_sampling import SMOTE, RandomOverSampler
from imblearn.under_sampling import RandomUnderSampler

from sklearn.metrics import roc_auc_score, RocCurveDisplay

# plt.style.use(('ggplot','tableau-colorblind10'))

from sklearn.inspection import permutation_importance



# Updating the function
def plot_numeric_vs_target(df, x, y, figsize=(6,4),
                           target_type='reg', estimator='mean',
                           errorbar='ci',sorted=False,ascending=False,
                           **kwargs): # kwargs for sns.regplot

  nulls = df[[x,y]].isna().sum()
  if nulls.sum()>0:
    print(f"- Excluding {nulls.sum()} NaN's")
    # print(nulls)
    temp_df = df.dropna(subset=[x,y,])
  else:
    temp_df = df
  # Create the figure
  fig, ax = plt.subplots(figsize=figsize)

  # REGRESSION-TARGET PLOT
  if 'reg' in target_type:
    # Calculate the correlation
    corr = df[[x,y]].corr().round(2)
    r = corr.loc[x,y]
    # Plot the data
    scatter_kws={'ec':'white','lw':1,'alpha':0.8}
    sns.regplot(data=temp_df, x=x, y=y, ax=ax, scatter_kws=scatter_kws, **kwargs) # Included the new argument within the sns.regplot function
    ## Add the title with the correlation
    ax.set_title(f"{x} vs. {y} (r = {r})")

  # CLASSIFICATION-TARGET PLOT
  elif 'class' in target_type:

    # Sort the groups by median/mean
    if sorted == True:

      if estimator == 'median':
        group_vals = temp_df.groupby(y)[x].median()
      elif estimator =='mean':
        group_vals = temp_df.groupby(y)[x].mean()

      ## Sort values
      group_vals = group_vals.sort_values(ascending=ascending)
      order = group_vals.index

    else:
      # Set order to None if not calcualted
      order = None


    # Left Subplot (barplot)
    sns.barplot(data=temp_df, x=y, y=x, order=order,  estimator=estimator,
                errorbar=errorbar, ax=ax, **kwargs)

    # Add title
    ax.set_title(f"{x} vs. {y}")

    # rotate xaxis labels
    ax.set_xticks(ax.get_xticks())
    ax.set_xticklabels(ax.get_xticklabels(), rotation=45, ha='right')


  # Final Adjustments & return
  fig.tight_layout()
  fig.show()
  return fig, ax


from sklearn.metrics import mean_absolute_error, mean_squared_error, r2_score
import pandas as pd


def regression_metrics(y_true, y_pred, label='', verbose = True, output_dict=False):
  # Get metrics
  mae = mean_absolute_error(y_true, y_pred)
  mse = mean_squared_error(y_true, y_pred)
  rmse = mean_squared_error(y_true, y_pred, squared=False)
  r_squared = r2_score(y_true, y_pred)
  if verbose == True:
    # Print Result with Label and Header
    header = "-"*60
    print(header, f"Regression Metrics: {label}", header, sep='\n')
    print(f"- MAE = {mae:,.3f}")
    print(f"- MSE = {mse:,.3f}")
    print(f"- RMSE = {rmse:,.3f}")
    print(f"- R^2 = {r_squared:,.3f}")
  if output_dict == True:
      metrics = {'Label':label, 'MAE':mae,
                 'MSE':mse, 'RMSE':rmse, 'R^2':r_squared}
      return metrics

def evaluate_regression(reg, X_train, y_train, X_test, y_test, verbose = True,
                        output_frame=False):
  # Get predictions for training data
  y_train_pred = reg.predict(X_train)

  # Call the helper function to obtain regression metrics for training data
  results_train = regression_metrics(y_train, y_train_pred, verbose = verbose,
                                     output_dict=output_frame,
                                     label='Training Data')
  print()
  # Get predictions for test data
  y_test_pred = reg.predict(X_test)
  # Call the helper function to obtain regression metrics for test data
  results_test = regression_metrics(y_test, y_test_pred, verbose = verbose,
                                  output_dict=output_frame,
                                    label='Test Data' )

  # Store results in a dataframe if ouput_frame is True
  if output_frame:
    results_df = pd.DataFrame([results_train,results_test])
    # Set the label as the index
    results_df = results_df.set_index('Label')
    # Set index.name to none to get a cleaner looking result
    results_df.index.name=None
    # Return the dataframe
    return results_df.round(3)




def overView(df):
  for key in df:
    print(f"Value Counts for '{key}'")
    print('*'*30)

    print(df[key].value_counts())
    print('*'*30)

    print(f"Value type is '{df[key].dtypes}'")
    print('*'*30)
    print(f"Value decription:- \n{df[key].describe()}")
    print('*'*30)

    print(f"Value unique = {round((df[key].nunique()/len(df)*100),2)}%")
    print('*'*30)

    print(f"Value NaN = {round((df[key].isna().sum()/len(df)*100), 2)}%")
    print('*'*30)

    print('\n')
    print('/'*40)